'Sort routines provided by John Petty

Type Sort_Order
    col As Long 'column to sort
    dir As Long '0 = ascending any other value = descending
End Type

Type Column_Type
    col As Long 'the column number
    tpe As Long 'type of data in column
    l As Long   'maximum length of data
    null As Long    'true if column contains nulls
End Type

Global ct() As Column_Type
Global so(), sotemp() As Sort_Order
Global CaseSensitive As Long

'Types of data
'   0 = undefined
'   1 = long
'   2 = double
'   3 = formatted currency string i.e $1,111.11
'   4 = date in format d/m/yyyy
'   5 = date in format m/d/yyyy
'   6 = date in format d-m-yyyy
'   7 = date in format m-d-yyyy
'   8 = string
'Note it can be possible to have a column of dates that can't be determined
'as to whether they are British or American format so either the locale should be obtained
'dates will accept a 2 digit year and assume in 21st century
'dates are not checked for validity
'or just the type for your country only included
'any column not clearly found will be treated as a string
'Note also doubles can't have exponent format

Sub XjpSetColTypes(dat() As String, c As Long, r As Long)
    Local x,y As Long 'loop counters
    ReDim ct(1 To c)
    For x = 1 To c
        ct(x).tpe = 8
        For y = 1 To r
           ct(x).l = Max(Len(dat(x, y)),ct(x).l)
        Next
        If ct(x).l Mod 4 <> 0 Then ct(x).l += 4 - (ct(x).l Mod 4)
    Next
End Sub

Sub jpGetColTypes(dat() As String, c As Long, r As Long)
    'dat is a multi dimensional array of the data in strings
    'all dimensions must be 1 based i.e 1 to x
    'c is number of cols 1 based
    'r is number of rows 1 based
    Local x As Long 'loop counter
    Local y As Long 'loop counter
    Local z As Long 'string position counter
    Local nd As Long 'not defined here
    Local t As String   'temp string so it can be changed
    Local s() As String 'for checking dates
    ReDim s(2)
    ReDim ct(1 To c)
    For x = 1 To c
        For y = 1 To r
'            If RTrim$(dat(x, y)) = "" Then  'check for a null field
            If RTrim$(dat(x, y)) = "" Or Left$(dat(x,y),1)= "^" Or Left$(dat(x,y),1)="#"  Then  'check for a null field
                ct(x).NULL = 1              'and set to true
            Else
                If Len(dat(x, y)) > ct(x).l Then ct(x).l = Len(dat(x, y)) 'get max length in column
                If ct(x).tpe < 3 Then   'check if column already contains not a long or double
                    nd = 0
                    t = Remove$(dat(x, y), ",")   'move dat to a temporary string without commas
                    For z = 1 To Len(t)
                        If InStr("0123456789.-", Mid$(t, z, 1)) = 0 Then
                            nd = -1
                            If ct(x).tpe = 1 Or ct(x).tpe = 2 Then ct(x).tpe = 0 'an entry in the column which isn't either of those
                            Exit For
                        End If
                    Next
                    If nd = 0 Then
                        If Tally(t, "-") > 1 Then
                            ct(x).tpe = 0   'could be a date
                            nd = -1
                        ElseIf Tally(t, "-") = 1 Then
                            If Left$(t, 1) <> "-" Then
                                ct(x).tpe = 8   'not a negative number so set as string
                                nd = -1
                            End If
                        End If
                    End If
                    If nd = 0 Then 'can be long or double
                        If InStr(t, ".") = 0 Then 'a long?
                            If ct(x).tpe = 0 Then ct(x).tpe = 1
                        Else
                            If Tally(t, ".") > 1 Then   'could be an IP address
                                ct(x).tpe = 8    'so set as string
                            Else
                                If ct(x).tpe < 2 Then ct(x).tpe = 3 'a long
                            End If
                        End If
                    End If
                End If
                If ct(x).tpe < 4 Then   'check for formatted currency
                    'could be improved to check for CR or DB etc so no negatve amounts at this time
                    nd = 0
                    t = Remove$(dat(x, y), ",")   'move dat to a temporary string without commas
                    If Left$(t, 1) = "$" Then
                        t = Right$(t, Len(t) - 1)
                        For z = 1 To Len(t)
                            If InStr("0123456789.", Mid$(t, z, 1)) = 0 Then
                                nd = -1
                                ct(x).tpe = 0   'not currency so undefined
                                Exit For
                            End If
                        Next
                        If nd = 0 Then ct(x).tpe = 3    'is currency
                    End If
                End If
                If ct(x).tpe = 0 Or (ct(x).tpe > 3 And ct(x).tpe < 8) Then  'check for dates
                    If Tally(dat(x, y), "-") = 2 Then
                        Reset s()
                        Parse dat(x, y), s(), "-"
                        If Val(s(0)) < 13 And Val(s(1)) > 12 Then   'is american format
                            ct(x).tpe = 7
                        ElseIf Val(s(0)) > 12 And Val(s(1)) < 13 Then   'is british format
                            ct(x).tpe = 6
                        Else
                            If ct(x).tpe <> 6 Then ct(x).tpe = 7   'default to american, change to suit
                        End If
                    ElseIf Tally(dat(x, y), "/") = 2 Then
                        Reset s()
                        Parse dat(x, y), s(), "/"
                        If Val(s(0)) < 13 And Val(s(1)) > 12 Then   'is american format
                            ct(x).tpe = 5
                        ElseIf Val(s(0)) > 12 And Val(s(1)) < 13 Then   'is british format
                            ct(x).tpe = 4
                        Else
                            If ct(x).tpe <> 4 Then ct(x).tpe = 5   'default to american, change to suit
                        End If
                    Else
                        ct(x).tpe = 8
                    End If
                End If
                If ct(x).tpe = 0 Then ct(x).tpe = 8 'none of above so a string
            End If
        Next
    Next
    For x = 1 To c
        If ct(x).tpe = 8 Then   'make string lengths a multiple of 4
            If ct(x).l Mod 4 <> 0 Then ct(x).l += 4 - (ct(x).l Mod 4)
        End If
    Next
End Sub

Sub jpSortCols(dat() As String, c As Long, r As Long, index() As Long, so() As Sort_Order)
    'GetColTypes must have been run once before any sorting
    'dat is a multi dimensional array of the data in strings
    'all dimensions must be 1 based i.e 1 to x
    'c is number of cols 1 based
    'r is number of rows 1 based
    'index() contains row numbers in original entry order i.e. 1 to r
    'and will return the display order after sorting
    'so() array of columns to be sorted, zero based array
    #Register None
    Local st As systemtime
    Local dbr As Double 'general purpose date
    Local v As Long 'general purpose long
    Local w As Long 'general purpose long
    Local x As Long 'loop counter
    Local y As Long 'loop counter
    Local z As Long 'string position counter
    Local sl As Long    'total length of data from a row to be sorted plus index
    Local bc As Long    'as counter for the byte array
    Local er As Long    'for date errors
    Local sp As String Pointer
    Local bp As Byte Ptr
    Local wp As Dword Ptr
    Local wp1 As Dword Ptr
    Local dp As Double Ptr
    Local t As String   'temp string
    Local s() As String 'for date parsing
    Local ind1P() As Byte Ptr   'array of pointers to each rows sort data
    Local first As Long
    Local last As Long
    Local stsize As Long    'recursion stack size
    Local stused As Long
    Local sortlen As Long   'length of sort data for each row in dwords
    Local cntr As Long
    Local hmem As Dword Ptr
    Local qstck() As Long    'recursion stack
    ReDim s(2)
    Local b() As Byte   'working space
    Local temp$
    '**** calculate working space needed and allocate ****
    For x = 0 To UBound(so)
        Select Case ct(so(x).col).tpe
            Case 8 'strings are most likely
                sl += ct(so(x).col).l
            Case 1  'longs
                sl += 4
            Case 2, 3   'doubles and currency
                sl += 8
            Case 4 To 7 'dates will be sorted as longs
                sl += 4
        End Select
        If ct(so(x).col).tpe <> 8 And ct(so(x).col).NULL Then sl += 4   'add space for the hidden null fields
    Next
    sl += 4 'and room for the indexes
    ReDim b(r * sl - 1)
    '**** work space allocated so now we load it ****
    For y = 1 To r
        For x = 0 To UBound(so)
            If ct(so(x).col).tpe <> 8 And ct(so(x).col).NULL Then'column has nulls so insert null sort field before
                wp = VarPtr(b(bc))      'the columns actual columnd data
                If RTrim$(dat(so(x).col, y)) = "" Then v = 0 Else v = 1
                If so(x).dir Then v = Not v 'descending
                @wp = v
                bc += 4
            End If
            Select Case ct(so(x).col).tpe
                Case 8 'strings are most likely
                    temp$ = IIf$(CaseSensitive, dat(so(x).col,y), UCase$(dat(so(x).col,y)))  '<--------------
                    'sp = StrPtr(dat(so(x).col, y))
                    sp = StrPtr(temp$)                                                       '<--------------
                    bp = sp
                    wp = VarPtr(b(bc))
                    If so(x).dir = 0 Then    'ascending
                        'For z = 0 To Len(dat(so(x).col, y)) - 1
                        For z = 0 To Len(temp$) - 1                                          '<--------------
                            b(bc) = @bp[z]
                            Incr bc
                        Next
                        For w = z To ct(so(x).col).L - 1
                            b(bc) = 0
                            Incr bc
                        Next
                    Else    'descending
                        'For z = 0 To Len(dat(so(x).col, y)) - 1
                        For z = 0 To Len(temp$) - 1                                          '<--------------
                            b(bc) = Not @bp[z]
                            Incr bc
                        Next
                        For w = z To ct(so(x).col).L - 1
                            b(bc) = &HFF
                            Incr bc
                        Next
                    End If
                    v = ct(so(x).col).L \ 4
                    !pushad     'and swap every 4 bytes to sort as dwords
                    !mov esi, wp
                    !mov ecx, v
                    swplp:
                    !mov eax, [esi] 'move 4 characters to eax
                    !bswap eax      'then swap them so they can be compared 4 at a time
                    !mov [esi], eax
                    !add esi, 4
                    !dec ecx
                    !jnz swplp
                    !popad
                Case 1  'longs
                    wp = VarPtr(b(bc))
                    t = Remove$(dat(so(x).col, y), ",")
                    v = Val(t)
                    Bit Toggle v, 31    'so we can sort as dwords
                    If so(x).dir Then v = Not v 'descending
                    @wp = v
                    bc += 4
                Case 2  'doubles
                    wp = VarPtr(b(bc))
                    t = Remove$(dat(so(x).col, y), ",")
                    dbr = Val(t)
                    dp = VarPtr(dbr)
                    wp1 = dp
                    If dbr < 0 Then
                        @wp1 = Not @wp1
                        @wp1[1] = Not @wp1[1]
                    Else
                        Bit Toggle @wp1, 31 'reverse sign bit
                    End If
                    If so(x).dir Then 'descending
                        @wp1 = Not @wp1
                        @wp1[1] = Not @wp1[1]
                    End If
                    @wp = @wp1[1]   'and load with dwords reversed
                    @wp[1] = @wp1
                    bc += 8
                Case 3  'currency but sort as double
                    wp = VarPtr(b(bc))
                    t = Remove$(dat(so(x).col, y), Any ",$")
                    dbr = Val(t)
                    dp = VarPtr(dbr)
                    wp1 = dp
                    If dbr < 0 Then
                        @wp1 = Not @wp1
                        @wp1[1] = Not @wp1[1]
                    Else
                        Bit Toggle @wp1, 31 'reverse sign bit
                    End If
                    If so(x).dir Then 'descending
                        @wp1 = Not @wp1
                        @wp1[1] = Not @wp1[1]
                    End If
                    @wp = @wp1[1]   'and load with dwords reversed
                    @wp[1] = @wp1
                    bc += 8
                Case 4  'date d/m/yyyy sort as long
                    Reset s()
                    wp = VarPtr(b(bc))
                    If ct(so(x).col).NULL And RTrim$(dat(so(x).col, y)) = "" Then
                        v = 0   'so we dont get errors trying to convert date to double
                    Else
                        Parse dat(so(x).col, y), s(), "/"
                        st.wday = Val(s(0))
                        st.wmonth = Val(s(1))
                        st.wyear = Val(s(2))
                        If st.wyear < 100 Then st.wyear += 2000
                        er = systemtimetovarianttime(st, dbr)
                        If er = 0 Then
                            ? "invalid date " & dat(so(x).col, y)
                            v = 0
                        Else
                            v = dbr
                        End If
                        Bit Toggle v, 31    'so we can sort as dwords
                        If so(x).dir Then v = Not v 'descending
                    End If
                    @wp = v
                    bc +=4
                Case 5  'date m/d/yyyy sort as long
                    Reset s()
                    wp = VarPtr(b(bc))
                    If ct(so(x).col).NULL And RTrim$(dat(so(x).col, y)) = "" Then
                        v = 0   'so we dont get errors trying to convert date to double
                    Else
                        Parse dat(so(x).col, y), s(), "/"
                        st.wday = Val(s(1))
                        st.wmonth = Val(s(0))
                        st.wyear = Val(s(2))
                        If st.wyear < 100 Then st.wyear += 2000
                        er = systemtimetovarianttime(st, dbr)
                        If er = 0 Then
                            ? "invalid date " & dat(so(x).col, y)
                            v = 0
                        Else
                            v = dbr
                        End If
                        Bit Toggle v, 31    'so we can sort as dwords
                        If so(x).dir Then v = Not v 'descending
                    End If
                    @wp = v
                    bc +=4
                Case 6  'date d-m-yyyy sort as long
                    Reset s()
                    wp = VarPtr(b(bc))
                    If ct(so(x).col).NULL And RTrim$(dat(so(x).col, y)) = "" Then
                        v = 0   'so we dont get errors trying to convert date to double
                    Else
                        Parse dat(so(x).col, y), s(), "-"
                        st.wday = Val(s(0))
                        st.wmonth = Val(s(1))
                        st.wyear = Val(s(2))
                        If st.wyear < 100 Then st.wyear += 2000
                        er = systemtimetovarianttime(st, dbr)
                        If er = 0 Then
                            ? "invalid date " & dat(so(x).col, y)
                            v = 0
                        Else
                            v = dbr
                        End If
                        Bit Toggle v, 31    'so we can sort as dwords
                        If so(x).dir Then v = Not v 'descending
                    End If
                    @wp = v
                    bc +=4
                Case 7  'date m-d-yyyy sort as long
                    Reset s()
                    wp = VarPtr(b(bc))
                    If ct(so(x).col).NULL And RTrim$(dat(so(x).col, y)) = "" Then
                        v = 0   'so we dont get errors trying to convert date to double
                    Else
                        Parse dat(so(x).col, y), s(), "-"
                        st.wday = Val(s(1))
                        st.wmonth = Val(s(0))
                        st.wyear = Val(s(2))
                        If st.wyear < 100 Then st.wyear += 2000
                        er = systemtimetovarianttime(st, dbr)
                        If er = 0 Then
                            ? "invalid date " & dat(so(x).col, y)
                            v = 0
                        Else
                            v = dbr
                        End If
                        Bit Toggle v, 31    'so we can sort as dwords
                        If so(x).dir Then v = Not v 'descending
                    End If
                    @wp = v
                    bc +=4
            End Select
        Next
        wp = VarPtr(b(bc))  'now add the index
        @wp = index(y)
        bc +=4
    Next
    '**** data loaded now sort it with a quicksort****
    'this is a modification of a sort posted by Steve Hutchison
    'fast but as only single core not superfat
    ReDim ind1p(r - 1)  'create array of pointers for each rows sort data so we never move data
    ind1p(0) = VarPtr(b(0))
    For x = 1 To UBound(ind1P)  'and load it
        ind1p(x) = ind1P(x - 1) + sl
    Next
    stsize = r^.5 + 10
    ReDim qstck(stsize)
    first = 0
    last = r - 1
    cntr = 0
    sortlen = sl \ 4
    hmem = VarPtr(qstck(0))
    wp1 = VarPtr(Ind1P(0))
        ! PUSHAD    'and start a quicksort
        ! MOV esi, wp1              ; source address IN ESI
        outer_loop:
            ' -------------------
            ' calculate midpoint
            ' -------------------
            ! MOV eax, last
            ! ADD eax, first
            ! SHR eax, 1
            ! MOV esi, wp1
            ' =========================
            ! MOV ebx, [esi+eax*4]      ; midpoint IN EBX
            ! mov wp, ebx
            ' =========================
            ! CLD
            ! MOV eax, first
            ! MOV edx, last
            ' ---------------------------------------------------------
            inner_loop:
            ! MOV esi, wp1
            ! MOV edi, ebx              'set EDI
            ! MOV esi, [esi + eax*4]    'set ESI
            ! MOV ecx, sortlen
            LessComp:
            ! CMPSD
            ! JA wl2
            ! JB lc1
            ! DEC ecx
            ! JNZ LessComp
            ! JMP wl2
            lc1:
            ! INC eax
            ! JMP inner_loop
            wl2:
            ! MOV esi, wp1
            ! MOV edi, ebx              'set EDI
            ! MOV esi, [esi + edx*4]    'set ESI
            ! MOV ecx, sortlen
            GreatComp:
            ! CMPSD
            ! JB wl2out
            ! JA gc1
            ! DEC ecx
            ! JNZ GreatComp
            ! JMP wl2out
            gc1:
            ! DEC edx
            ! JMP wl2
            wl2Out:
            ! MOV esi, wp1
            ! CMP eax, edx              ; IF eax > edx, EXIT inner LOOP
            ! JG exit_innerx
            ' =========================
            ! MOV ecx, [esi+eax*4]
            ! MOV ebx, [esi+edx*4]      ; SWAP elements (pointers)
            ! MOV [esi+eax*4], ebx
            ! MOV [esi+edx*4], ecx
            ' =========================
            ! mov ebx, wp
            ! INC eax
            ! DEC edx
            ! CMP eax, edx
            ! JLE inner_loop
            exit_innerx:
            ' ---------------------------------------------------------
            ! CMP eax, last             ; IF eax < Last jump over
            ! JG iNxt
            ' =========================
            ! MOV edi, hmem
            ! MOV ecx, cntr
            ! MOV [edi+ecx*4], eax
            ! MOV ebx, last
            ! MOV [edi+ecx*4+4], ebx
            ' =========================
            ! ADD cntr, 2
            ! ADD ecx, 2
            ! cmp ecx, stused  ;if current stack size biggest
            ! JS iNxt
            ! MOV stused, ecx  ;save for logging
            iNxt:
            ! MOV last, edx             ; Last  = EDX
            ! CMP edx, first            ; compare Last & First
            ! JG outer_loop

        ! CMP cntr, 0
        ! JZ qsOut
        ! SUB cntr, 2
        ' =========================
        ! MOV edi, hmem
        ! MOV ecx, cntr
        ! MOV ebx, [edi+ecx*4]
        ! MOV first, ebx
        ! MOV ebx, [edi+ecx*4+4]
        ! MOV last, ebx
        ' =========================
        ! JMP outer_loop
        qsOut:
        ! POPAD
    '**** sort finshed so put result in index ****
    For x = 1 To r
        wp = ind1p(x - 1) + sl - 4
        index(x) = @wp
    Next
End Sub
